# =============================================================================
# >> CONSTANTS
# =============================================================================
SKIP_OPTIONAL_CHECK = False


# =============================================================================
# >> CLASSES
# =============================================================================
class BaseStruct(dict):
    pass


class StructField:
    def __init__(self, attr_name, field_name, value_type, optional=False):
        self.attr_name = attr_name
        self.field_name = field_name
        self.value_type = value_type
        self.optional = optional


class StructMeta(type):
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)

        cls._fields_by_attr_name = {}
        cls._fields_by_field_name = {}

        for field in cls._fields:
            cls._fields_by_attr_name[field.attr_name] = field
            cls._fields_by_field_name[field.field_name] = field


class Struct(BaseStruct, metaclass=StructMeta):
    _fields = ()

    def __init__(self, *args, **kwargs):
        super().__init__()

        if len(args) == 1 and not kwargs:

            # Initialization from a unJSONed dict with OBS naming
            for field_name, raw_value in args[0].items():
                try:
                    field = self._fields_by_field_name[field_name]
                except KeyError:
                    raise TypeError(
                        "Struct '{class_name}': '{field_name}' is an invalid "
                        "field name for this struct".format(
                            class_name=self.__class__.__name__,
                            field_name=field_name))

                value = field.value_type(raw_value)
                self[field_name] = value

        elif not args:

            # Initialization from Python-friendly attributes
            for attr_name, value in kwargs.items():
                try:
                    field = self._fields_by_attr_name[attr_name]
                except KeyError:
                    raise TypeError(
                        "Struct '{class_name}': '{attr_name}' is an invalid "
                        "keyword argument for this struct".format(
                            class_name=self.__class__.__name__,
                            attr_name=attr_name))

                if not isinstance(value, field.value_type):
                    raise TypeError(
                        "Struct '{class_name}': '{attr_name}' value should be "
                        "of type '{type_name}'".format(
                            class_name=self.__class__.__name__,
                            attr_name=attr_name,
                            type_name=field.value_type.__name__))

                self[field.field_name] = value

        else:
            raise TypeError(
                "{class_name} can initialized only either with 1 positional "
                "argument (dictionary of fields) or with keyword "
                "arguments".format(class_name=self.__class__.__name__))

        # Make sure that all required fields are set
        for field in self._fields:
            if SKIP_OPTIONAL_CHECK or field.optional:
                continue

            if field.field_name not in self:
                raise TypeError(
                    "Struct '{class_name}': Missing required {field_name} "
                    "field".format(
                        class_name=self.__class__.__name__,
                        field_name=field.field_name))

    def __getattr__(self, attr_name):
        try:
            field = self._fields_by_attr_name[attr_name]
        except KeyError:
            raise AttributeError(
                "'{class_name}' object has no attribute '{attr_name}'".format(
                    class_name=self.__class__.__name__, attr_name=attr_name))

        return self[field.field_name]

    def __setattr__(self, attr_name, value):
        try:
            field = self._fields_by_attr_name[attr_name]
        except KeyError:
            raise AttributeError(
                "'{class_name}' object has no attribute '{attr_name}'".format(
                    class_name=self.__class__.__name__, attr_name=attr_name))

        if not isinstance(value, field.value_type):
            raise TypeError(
                "Struct '{class_name}': '{attr_name}' value should be of type "
                "'{type_name}'".format(
                    class_name=self.__class__.__name__,
                    attr_name=attr_name,
                    type_name=field.value_type.__name__))

        self[field.field_name] = value


class VariableStruct(BaseStruct):
    _allowed_types = ()

    def __init__(self, **kwargs):
        super().__init__()

        # Initialization from OBS-named kwargs
        for field_name, value in kwargs.items():
            if type(value) not in self._allowed_types:
                raise TypeError(
                    "VariableStruct '{class_name}': '{field_name}' value "
                    "should be one of these types: {types}".format(
                        class_name=self.__class__.__name__,
                        field_name=field_name,
                        types=', '.join(
                            type_.__name__ for type_ in self._allowed_types
                        )
                    )
                )

            self[field_name] = value

        return

    def __getattr__(self, field_name):
        return self[field_name]

    def __setattr__(self, field_name, value):
        if type(value) not in self._allowed_types:
            raise TypeError(
                "VariableStruct '{class_name}': '{field_name}' value should "
                "be one of these types: {types}".format(
                    class_name=self.__class__.__name__,
                    field_name=field_name,
                    types=', '.join(
                        type_.__name__ for type_ in self._allowed_types
                    )
                )
            )

        self[field_name] = value



"""
This module holds dynamically generated classes.
For more info see protocol.py and protocol.json.
"""

# =============================================================================
# >> IMPORTS
# =============================================================================
# obs-ws-rc
from .struct import Struct, StructField, StructMeta


# =============================================================================
# >> GLOBAL VARIABLES
# =============================================================================
# All classes that inherit from BaseEvent will be available by their names
# in this dictionary
events = {}


# =============================================================================
# >> BASE CLASSES
# =============================================================================
class BaseEventMeta(StructMeta):
    def __init__(cls, name, bases, namespace):
        cls._fields = cls._fields[:] + (
            StructField('update_type', "update-type", str),
            StructField('stream_timecode', "stream-timecode", str, True),
            StructField('rec_timecode', "rec-timecode", str, True),
        )

        super().__init__(name, bases, namespace)

        try:
            type_name = cls.type_name
        except NotImplementedError:
            return

        if type_name in events:
            raise ValueError(
                "Class for event '{type_name}' already exists".format(
                    type_name=type_name))

        events[type_name] = cls


class BaseEvent(Struct, metaclass=BaseEventMeta):
    @property
    def type_name(self):
        raise NotImplementedError



"""
This module holds dynamically generated classes.
For more info see protocol.py and protocol.json.
"""

# =============================================================================
# >> IMPORTS
# =============================================================================
# Python
from enum import Enum

# obs-ws-rc
from .struct import Struct, StructField, StructMeta


# =============================================================================
# >> BASE CLASSES
# =============================================================================
class ResponseStatus(Enum):
    OK = 'OK'
    ERROR = 'ERROR'


class BaseResponseMeta(StructMeta):
    def __init__(cls, name, bases, namespace):
        cls._fields = cls._fields[:] + (
            StructField('message_id', "message-id", str),
            StructField(
                'status',
                "status",
                lambda status: ResponseStatus(status.upper())
            ),
            StructField('error', "error", str, True),
        )

        super().__init__(name, bases, namespace)


class BaseResponse(Struct, metaclass=BaseResponseMeta):
    pass


class BaseRequest(Struct):
    @property
    def type_name(self):
        raise NotImplementedError

    class response_class(BaseResponse):
        pass

    def get_request_data(self, message_id):
        dict_ = self.copy()
        dict_['request-type'] = self.type_name
        dict_['message-id'] = message_id
        return dict_


def dummy_request(**kwargs):
    raise NotImplementedError("protocol.json doesn't implement this request")

AuthenticateRequest = dummy_request
GetAuthRequiredRequest = dummy_request



# =============================================================================
# >> IMPORTS
# =============================================================================
# obs-ws-rc
from .events import BaseEvent
from .requests import BaseRequest, BaseResponse
from .struct import Struct, StructField, VariableStruct


# =============================================================================
# >> FUNCTIONS
# =============================================================================
def _get_fields(obj, known_types):
    fields = []
    for field_obj in obj.get('fields', []):
        attr_name = field_obj['attr_name']
        field_name = field_obj['field_name']
        value_type_str = field_obj['type']
        try:
            value_type = known_types[value_type_str]
        except KeyError:
            raise UnknownType(value_type_str)

        optional = field_obj.get('optional', False)

        fields.append(StructField(
            attr_name, field_name, value_type, optional))

    return tuple(fields)


def _get_allowed_types(obj, known_types):
    allowed_types = []

    for allowed_type_str in obj['allowed_types']:
        try:
            allowed_types.append(known_types[allowed_type_str])
        except KeyError:
            raise UnknownType(allowed_type_str)

    return tuple(allowed_types)


def _get_item_type(obj, known_types):
    item_type_str = obj['item_type']
    try:
        return known_types[item_type_str]
    except KeyError:
        raise UnknownType(item_type_str)


def build_types(protocol):
    struct_meta = type(Struct)
    var_struct_meta = type(VariableStruct)

    known_types = {
        'str': str,
        'int': int,
        'float': float,
        'bool': bool
    }

    types_to_build = list(protocol['types'].items())
    progress = True
    while types_to_build and progress:
        progress = False

        for type_name, type_obj in tuple(types_to_build):
            if type_obj['type'] == "struct":
                try:
                    type_fields = _get_fields(type_obj, known_types)
                except UnknownType:
                    continue

                type_class = struct_meta(
                    type_name,
                    (Struct, ),
                    {'_fields': type_fields}
                )

            elif type_obj['type'] == "var_struct":
                try:
                    allowed_types = _get_allowed_types(type_obj, known_types)
                except UnknownType:
                    continue

                type_class = var_struct_meta(
                    type_name,
                    (VariableStruct, ),
                    {'_allowed_types': allowed_types}
                )

            elif type_obj['type'] == "collection":
                try:
                    item_type = _get_item_type(type_obj, known_types)
                except UnknownType:
                    continue

                def get_type_class(item_type):
                    return lambda data: list(map(item_type, data))

                type_class = get_type_class(item_type)

            else:
                raise ValueError("Unknown type type '{type_name}'".format(
                    type_name=type_name))

            known_types[type_name] = type_class
            types_to_build.remove((type_name, type_obj))
            progress = True

    if types_to_build:
        raise ValueError(
            "The following types have unresolved dependencies: {types}".format(
                types=', '.join(
                    type_name for type_name, type_obj in types_to_build)))

    return known_types


def build_requests(protocol, known_types):
    response_meta = type(BaseResponse)
    request_meta = type(BaseRequest)

    result = {}

    for request_name, request_obj in protocol['requests'].items():
        request_class_name = request_name + "Request"
        response_obj = request_obj.get('response', {})

        request_fields = _get_fields(request_obj, known_types)
        response_fields = _get_fields(response_obj, known_types)

        response_class = response_meta(
            'response_class',
            (BaseResponse, ),
            {'_fields': response_fields}
        )

        request_class = request_meta(
            request_class_name,
            (BaseRequest, ),
            {
                '_fields': request_fields,
                'response_class': response_class,
                'type_name': request_name
            }
        )

        result[request_class_name] = request_class

    return result


def build_events(protocol, known_types):
    event_meta = type(BaseEvent)

    result = {}

    for event_name, event_obj in protocol['events'].items():
        event_class_name = event_name + "Event"

        event_fields = _get_fields(event_obj, known_types)

        event_class = event_meta(
            event_class_name,
            (BaseEvent, ),
            {'_fields': event_fields, 'type_name': event_name}
        )

        result[event_class_name] = event_class

    return result


# =============================================================================
# >> CLASSES
# =============================================================================
class UnknownType(Exception):
    pass



# =============================================================================
# >> IMPORTS
# =============================================================================
# Python
import asyncio
from base64 import b64encode
from hashlib import sha256
import json
from traceback import format_exc

# websockets
import websockets

# obs-ws-rc
from .events import events
from .logs import logger
from .requests import (
    AuthenticateRequest, GetAuthRequiredRequest, ResponseStatus)


# =============================================================================
# >> CONSTANTS
# =============================================================================
URI_TEMPLATE = "ws://{host}:{port}/"
DEFAULT_PORT = 4444


# =============================================================================
# >> CLASSES
# =============================================================================
class AuthError(Exception):
    """Raised by :meth:`OBSWS.connect` if authentication has failed."""
    pass


class OBSWS:
    """Main class used for obs-websocket communication. Can be used as a
    context manager (given you use it in ``async with`` statement).

    Example usage::

        async with OBSWS("localhost") as obsws:
            ...

    This is an equivalent to the following::

        obsws = OBSWS("localhost")
        await obsws.connect()

        try:
            ...
        finally:
            await obsws.close()

    This class also supports Future-like protocol (it implements
    :meth:`__await__` method). You can await for the OBSWS instance for it
    to close::

        await obsws

    .. note::
        When entering the context manager (using ``async with`` statement),
        you should be ready to except :exc:`AuthError` that might raise due to
        failed auth, or :exc:`OSError` that can be raised by the underlying
        websockets library in case of being unable to connect to OBS Studio.

    .. seealso::
        :meth:`connect`
        :meth:`close`

    """
    def __init__(self, host, port=DEFAULT_PORT, password=None, *,
                 skip_auth=False, loop=None):

        """
        :param str host: Server host
        :param int port: Server port
        :param str|None password: Server password (if needed)
        :param bool skip_auth: Whether or not to skip authentication
        :param asyncio.AbstractEventLoop|None loop: Event loop to use

        """
        self._host = host
        self._port = port
        self._password = password
        self._skip_auth = skip_auth

        if loop is None:
            loop = asyncio.get_event_loop()

        self._loop = loop

        self._message_map = {}
        self._message_count = 0

        self._event_handlers = {}

        self._ws = None
        self._ws_close_event = asyncio.Event(loop=self._loop)
        self._done_event = asyncio.Event(loop=self._loop)
        self._done_event.set()
        self._event_futures = {}

    def __await__(self):
        async def coro():
            return await self._done_event.wait()

        return coro().__await__()

    @property
    def host(self):
        """The host that OBSWS was instantiated with (read-only).

        :return: Server host
        :rtype: str

        """
        return self._host

    @property
    def port(self):
        """The port that OBSWS was instantiated with (read-only).

        :return: Server port
        :rtype: int

        """
        return self._port

    @property
    def password(self):
        """The port that OBSWS was instantiated with (read-only).

        :return: Server password (``None`` if not given)
        :rtype: str|None

        """
        return self._password

    @property
    def closed(self):
        """Return whether or not this OBSWS instance is closed."""
        return self._done_event.is_set()

    async def connect(self):
        """Establish connection to the server, start the event loop and
        perform authentication (the latter can be skipped with ``skip_auth``
        argument in :meth:`__init__`).

        :raises ValueError: if already connected
        :raises AuthError: if auth is enabled but password is invalid or not
                           not set
        :raises OSError: raised by the underlying websockets library if
                         connection attempt is failed

        .. note::
            This method is a coroutine.

        """
        if self._ws is not None:
            raise ValueError("Already connected")

        self._ws = await websockets.connect(URI_TEMPLATE.format(
            host=self._host,
            port=self._port
        ))
        self._ws_close_event.clear()

        asyncio.ensure_future(self._recv_loop())

        if not self._skip_auth:
            try:
                await self._authenticate()
            except AuthError:
                await self._close()
                await self._done_event.wait()
                raise

    async def _close(self):
        """Close the underlying websocket connection.

        .. note::
            This method is a coroutine.

        """
        if self._ws is None:
            return

        try:
            await self._ws.close()
        except ConnectionResetError:
            pass

        self._ws = None
        self._ws_close_event.set()

        for future in self._message_map.values():
            future.set_result(None)

        for future in self._event_futures.values():
            future.set_result(None)

        self._message_map.clear()

    async def close(self):
        """Clean shutdown. Consequent calls on an already closed instance have
        not effect.

        .. note::
            This method is a coroutine.

        """
        await self._close()
        await self._done_event.wait()

    async def __aenter__(self):
        """Enter context: connect and return self.

        :return: Ready-to-work OBSWS instance
        :rtype: OBSWS

        .. note::
            This method is a coroutine.

        """
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Leave context: clean shutdown.

        .. note::
            This method is a coroutine.

        """
        await self._close()
        await self._done_event.wait()

    async def _recv_loop(self):
        self._done_event.clear()

        while not self._ws_close_event.is_set():
            try:
                data = json.loads(await self._ws.recv())
            except websockets.ConnectionClosed:
                await self._close()

            else:
                message_id = data.get('message-id')
                if message_id is not None:
                    self._message_map.pop(message_id).set_result(data)
                    continue

                type_name = data.get('update-type')
                if type_name is not None:
                    asyncio.ensure_future(
                        self._handle_event(type_name, data), loop=self._loop)

                    continue

                # TODO: Not a response nor an event - log an error maybe?

        self._done_event.set()

    async def _authenticate(self):
        response = await self.require(GetAuthRequiredRequest())
        if not response.auth_required:
            return

        if self._password is None:
            raise AuthError(
                "Auth is required, but the password is not provided")

        auth_string = b64encode(sha256(b64encode(
            sha256((self._password + response.salt).encode('utf-8')).digest()
        ) + response.challenge.encode('utf-8')).digest()).decode('ascii')

        response = await self.require(AuthenticateRequest(auth=auth_string))
        if response.status == ResponseStatus.ERROR:
            raise AuthError(
                "Auth has failed, error from the server: '{error}'".format(
                    error=response.error))

    async def require(self, request):
        """Send a request to the server and await, return the response.

        :param requests.BaseRequest request: Fully formed request
        :return: Response from the server (None if the connection was
                 closed during communication)
        :rtype: requests.BaseResponse|None
        :raises ValueError: if not connected

        .. note::
            This method is a coroutine.

        """
        if self._ws is None:
            raise ValueError("Not connected")

        self._message_count += 1
        message_id = str(self._message_count)
        future = self._message_map[message_id] = self._loop.create_future()

        data = request.get_request_data(message_id)
        await self._ws.send(json.dumps(data))

        message = await future
        if message is None:
            return None

        return request.response_class(message)

    async def event(self, type_name=None):
        """Await for an event of type ``type_name``. If ``type_name`` is None,
        await for any event. Return the event.

        :param str|None type_name: Event type to await for, ``None`` to await
                                   for an event of any type
        :return: Event (None if the connection was closed during awaiting)
        :rtype: events.BaseEvent|None
        :raises ValueError: if not connected

        .. note::
            This method is a coroutine.

        """
        if self._ws is None:
            raise ValueError("Not connected")

        if type_name not in self._event_futures:
            self._event_futures[type_name] = self._loop.create_future()

        return await self._event_futures[type_name]

    async def _handle_event(self, type_name, data):
        event_class = events.get(type_name)

        # Do we know of this event type?
        if event_class is None:

            # Not implemented in our protocol, no further action required
            return

        callbacks = self._event_handlers.get(type_name)
        future_any = self._event_futures.pop(None, None)
        future_event = self._event_futures.pop(type_name, None)

        # Is there anybody willing to handle it?
        if callbacks is None and future_any is None and future_event is None:

            # No, so we don't even instantiate the event class
            return

        try:
            event = event_class(data)
        except:
            logger.error(
                "OBS-WS-RC: '{type_name}' event instantiation raised (invalid "
                "protocol.json?)\n{exc}\nThe callbacks are not going to be "
                "called!".format(type_name=type_name, exc=format_exc()))

            return

        if future_any is not None:
            future_any.set_result(event)

        if future_event is not None:
            future_event.set_result(event)

        if callbacks is not None:
            coros = []
            for callback in callbacks:
                if asyncio.iscoroutinefunction(callback):
                    coros.append(callback(self, event))
                else:
                    try:
                        callback(self, event)
                    except:
                        logger.error(
                            "OBS-WS-RC: '{type_name}' event "
                            "handler raised!\n{exc}\n".format(
                                type_name=type_name, exc=format_exc()))

            close_future = asyncio.ensure_future(self._ws_close_event.wait(),
                                                 loop=self._loop)

            gather_future = asyncio.gather(
                *coros, return_exceptions=True, loop=self._loop)

            done, pending = await asyncio.wait(
                [close_future, gather_future],
                return_when=asyncio.FIRST_COMPLETED,
                loop=self._loop
            )

            if gather_future.done():
                results = await gather_future

                for result in results:
                    if isinstance(result, BaseException):
                        try:
                            raise result
                        except:
                            logger.error(
                                "OBS-WS-RC: '{type_name}' event "
                                "async handler raised!\n{exc}\n".format(
                                    type_name=type_name, exc=format_exc()))

            else:
                gather_future.cancel()

            if not close_future.done():
                close_future.cancel()

    def register_event_handler(self, type_name, callback):
        """Register event handler (either a regular one or an async-coroutine).

        :param type_name: Event name
        :param callable callback: Function or coroutine function
        :raises ValueError: if callback is already registered for the event

        .. deprecated:: 2.2
            Use :meth:`event` instead.

        """
        if type_name not in self._event_handlers:
            self._event_handlers[type_name] = []

        if callback in self._event_handlers[type_name]:
            raise ValueError(
                "Callback '{callback}' is already registered to handle "
                "'{type_name}' event".format(
                    callback=callback, type_name=type_name))

        self._event_handlers[type_name].append(callback)

    def unregister_event_handler(self, type_name, callback):
        """Unregister previously registered event handler.

        :param type_name: Event name
        :param callable callback: Function or coroutine function

        .. deprecated:: 2.2
            Use :meth:`event` instead.

        """
        self._event_handlers[type_name].remove(callback)
        if not self._event_handlers[type_name]:
            del self._event_handlers[type_name]



from pathlib import Path
from setuptools import setup


root = Path(__file__).parent.absolute()

with open(str(root / 'README.rst')) as f:
    long_description = f.read()


setup(
    name='obs-ws-rc',

    version='2.2.0',

    description=("asyncio-based Python 3.5+ client to obs-websocket "
                 "plugin for OBS Studio"),

    long_description=long_description,

    url="https://github.com/KirillMysnik/obs-ws-rc",

    author="Kirill Mysnik",

    author_email = "kirill@mysnik.com",

    license='MIT',

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Multimedia :: Video',

        'License :: OSI Approved :: MIT License',

        'Operating System :: OS Independent',

        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='obs obs-websocket',

    packages=['obswsrc', ],

    install_requires=['websockets', ],

    python_requires='>=3.5',

    package_data={
        'obswsrc': ["protocol.json", ],
    }
)
